// [Config File Handler] - Handle connections to config file

#include "configfile_handler.h"
#include "functions.h"
#include <fstream>
#include <sstream>
#include <algorithm>
#include <iostream>
#include <stdexcept>

std::map<std::string, std::map<std::string, std::string>> ConfigFileHandler::configData;

// Check if file exist
bool ConfigFileHandler::fileExists(const std::string& filePath) {
    std::ifstream file(filePath);
    return file.good();
}

bool ConfigFileHandler::loadConfig(const std::string& filePath) {
    std::ifstream file(filePath);
    if (!file.is_open()) {
        return false;
    }

    std::string line, section;
    while (std::getline(file, line)) {
        line = normalizeLineEndings(line);
        line = trim(line);

        // Skip empty rows and comments
        if (line.empty() || line[0] == ';' || line[0] == '#') {
            continue;
        }

        // Check for section headers
        if (line[0] == '[' && line.back() == ']') {
            section = line.substr(1, line.size() - 2);
        } else {
            // Parse Key-Value pairs
            size_t pos = line.find('=');
            if (pos != std::string::npos && !section.empty()) {
                std::string key = trim(line.substr(0, pos));
                std::string value = trim(line.substr(pos + 1));
                configData[section][key] = value;
            }
        }
    }
    file.close();
    return true;
}

// Get string from config file
std::string ConfigFileHandler::getParam(const std::string& section, const std::string& key) {
    if (configData.count(section)) {
        if (configData[section].count(key)) {
            return configData[section][key];
        } else {
            std::string fullLogMessage = "Key not found: " + std::string(key);
            logMessage(LogLevel::Error, fullLogMessage);
        }
    } else {
        std::string fullLogMessage = "Section not found: " + std::string(section);
        logMessage(LogLevel::Error, fullLogMessage);
    }
    return "";
}

// Get integer  from config file
int ConfigFileHandler::getIntParam(const std::string& section, const std::string& key) {
    std::string value = getParam(section, key);

    if (value.empty()) {
        std::string fullLogMessage = "Key not found or empty value for: [" + std::string(section) + "] " + std::string(key);
        logMessage(LogLevel::Error, fullLogMessage);
        throw std::runtime_error("Missing or empty value for key: " + key);
    }

    try {
        return std::stoi(value);
    } catch (const std::invalid_argument&) {
        std::string fullLogMessage = "Invalid integer format for key: [" + std::string(section) + "] " + std::string(key) + " with value: " + std::string(value);
        logMessage(LogLevel::Error, fullLogMessage);
        throw std::runtime_error("Invalid integer value for key: " + key);
    } catch (const std::out_of_range&) {
        std::string fullLogMessage = "Integer value out of range for key: [" + std::string(section) + "] " + std::string(key) + " with value: " + std::string(value);
        logMessage(LogLevel::Error, fullLogMessage);
        throw std::runtime_error("Integer value out of range for key: " + key);
    }
}

std::string ConfigFileHandler::trim(const std::string& str) {
    size_t first = str.find_first_not_of(" \t");
    if (first == std::string::npos) {
        return "";
    }
    size_t last = str.find_last_not_of(" \t");
    return str.substr(first, (last - first + 1));
}

// Remove Windows Carriage-Returns (\r)
std::string ConfigFileHandler::normalizeLineEndings(const std::string& line) {
    std::string result = line;
    result.erase(std::remove(result.begin(), result.end(), '\r'), result.end());
    return result;
}
