// [Database Handler] - Handle connections to database server

#include "database_handler.h"
#include "functions.h"
#include <iostream>

bool check_db_connection(const std::string& server, const std::string& user,
                         const std::string& password, const std::string& dbname,
                         unsigned int port, MYSQL*& conn) {
    conn = mysql_init(nullptr);
    if (!conn) {
        logMessage(LogLevel::Error, "Error: Failed to initialize MySQL handler");
        return false;
    }

    // Require TCP-Connection
    int protocol = MYSQL_PROTOCOL_TCP;
    if (mysql_options(conn, MYSQL_OPT_PROTOCOL, &protocol)) {
        logMessage(LogLevel::Error, "Error: Failed to set protocol option");
        mysql_close(conn);
        return false;
    }

    if (!mysql_real_connect(conn, server.c_str(), user.c_str(), password.c_str(),
                            dbname.c_str(), port, nullptr, 0)) {
        std::string fullLogMessage = "DB Connection Error: " + std::string(mysql_error(conn));
        logMessage(LogLevel::Error, fullLogMessage);
        mysql_close(conn);
        return false;
    }

    logMessage(LogLevel::Verbose, "Connected successfully to the database!");
    return true;
}

//Check if table 'devicelog' exists, if not than create this table
bool check_table_exists(MYSQL* conn, const std::string& dbname) {
    std::string checkTableQuery =
        "SELECT COUNT(*) FROM information_schema.tables WHERE table_schema = '" + dbname + "' AND table_name = 'devicelog';";

    if (mysql_query(conn, checkTableQuery.c_str())) {
        std::string fullLogMessage = "Error checking table existence: " + std::string(mysql_error(conn));
        logMessage(LogLevel::Error, fullLogMessage);
        return false;
    }

    MYSQL_RES* result = mysql_store_result(conn);
    if (!result) {
        std::string fullLogMessage = "Error retrieving check table result: " + std::string(mysql_error(conn));
        logMessage(LogLevel::Error, fullLogMessage);
        return false;
    }

    MYSQL_ROW row = mysql_fetch_row(result);
    bool tableExists = (row && std::stoi(row[0]) > 0);
    mysql_free_result(result);

    if (!tableExists) {
        logMessage(LogLevel::Verbose, "Table 'devicelog' does not exist. Creating the table...");

        const char* createTableQuery = R"(
            CREATE TABLE devicelog (
                id INT(11) AUTO_INCREMENT PRIMARY KEY,
                datetime DATETIME NOT NULL,
                device_addr TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NOT NULL,
                network TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NOT NULL,
                fcnt TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NOT NULL,
                packet_counter INT NOT NULL,
                dir TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NOT NULL,
                rssi FLOAT NOT NULL,
                snr FLOAT NOT NULL,
                gateway_eui TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NOT NULL,
                payload TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NOT NULL
            );
        )";

        if (mysql_query(conn, createTableQuery)) {
            std::string fullLogMessage = "Error creating table 'devicelog':" + std::string(mysql_error(conn));
            logMessage(LogLevel::Error, fullLogMessage);
            return false;
        }
        logMessage(LogLevel::Normal, "Table 'devicelog' created successfully!");
    } else {
        logMessage(LogLevel::Verbose, "Table 'devicelog' already exists.");
    }

    return true;
}

// Check if devaddr exist not in table and insert it. If devaddr exists, update entry.
bool db_insert_or_update_packet(MYSQL* conn, const std::string& datetime, const std::string& device_addr,
                                 const std::string& network, const std::string& fcnt, const std::string& dir,
                                 float rssi, float snr, const std::string& gateway_eui, const std::string& payload) {

    std::string checkQuery = "SELECT fcnt FROM devicelog WHERE device_addr = '" + device_addr + "';";

    if (mysql_query(conn, checkQuery.c_str())) {
        std::string fullLogMessage = "Error checking for existing device_addr: " + std::string(mysql_error(conn));
        logMessage(LogLevel::Error, fullLogMessage);
        mysql_close(conn);
        return false;
    }

    MYSQL_RES* res = mysql_store_result(conn);
    if (!res) {
        std::string fullMessage = "Error storing result: " + std::string(mysql_error(conn));
        logMessage(LogLevel::Error, fullMessage);
        mysql_close(conn);
        return false;
    }

    MYSQL_ROW row = mysql_fetch_row(res);
    bool exists = (row != nullptr);
    bool fcntMatches = exists && (row[0] != nullptr && fcnt == row[0]);
    mysql_free_result(res);

    if (exists) {
        std::string updateQuery;
        if (fcntMatches) {
            updateQuery = "UPDATE devicelog SET "
                          "datetime = '" + datetime + "', "
                          "network = '" + network + "', "
                          "fcnt = '" + fcnt + "', "
                          "dir = '" + dir + "', "
                          "rssi = " + std::to_string(rssi) + ", "
                          "snr = " + std::to_string(snr) + ", "
                          "gateway_eui = '" + gateway_eui + "', "
                          "payload = '" + payload + "' "
                          "WHERE device_addr = '" + device_addr + "';";
        } else {
            updateQuery = "UPDATE devicelog SET "
                          "datetime = '" + datetime + "', "
                          "network = '" + network + "', "
                          "fcnt = '" + fcnt + "', "
                          "packet_counter = packet_counter + 1, "
                          "dir = '" + dir + "', "
                          "rssi = " + std::to_string(rssi) + ", "
                          "snr = " + std::to_string(snr) + ", "
                          "gateway_eui = '" + gateway_eui + "', "
                          "payload = '" + payload + "' "
                          "WHERE device_addr = '" + device_addr + "';";
        }

        if (mysql_query(conn, updateQuery.c_str())) {
            std::string fullLogMessage = "Error updating Device Log: " + std::string(mysql_error(conn));
            logMessage(LogLevel::Error, fullLogMessage);
            mysql_close(conn);
            return false;
        }

        logMessage(LogLevel::Verbose, fcntMatches ? "Update Device Log without incrementing packet counter." : "Update Device Log with incremented packet counter.");
    } else {
        std::string insertQuery = "INSERT INTO devicelog (datetime, device_addr, network, fcnt, packet_counter, dir, rssi, snr, gateway_eui, payload) "
                                  "VALUES ('" + datetime + "', '" + device_addr + "', '" + network + "', '" + fcnt + "', 1, '" +
                                  dir + "', " + std::to_string(rssi) + ", " + std::to_string(snr) + ", '" + gateway_eui + "', '" + payload + "');";

        if (mysql_query(conn, insertQuery.c_str())) {
            std::string fullLogMessage = "Error inserting Device Log: " + std::string(mysql_error(conn));
            logMessage(LogLevel::Error, fullLogMessage);
            mysql_close(conn);
            return false;
        }

        logMessage(LogLevel::Verbose, "Insert Device Log in Database.");
    }

    mysql_close(conn);
    return true;
}
