// [Functions] - Functions like string of current time, logs, extracting JSON, ...

#include "functions.h"

#include <iostream>
#include <string>
#include <cstring>
#include <iomanip>
#include <chrono>
#include <ctime>
#include <map>
#include <sstream>
#include <vector>
#include <stdexcept>

// Static variable to store the current log level
static LogLevel currentLogLevel = LogLevel::Normal;

// Set log level
void setLogLevel(int level) {
    if (level < 0 || level > 3) {
        throw std::invalid_argument("Log level must be between 0 (Silent) and 3 (Verbose).");
    }
    currentLogLevel = static_cast<LogLevel>(level);
}

// Print Logs & Error Information
void logMessage(LogLevel level, const std::string& message) {
    if (static_cast<int>(level) <= static_cast<int>(currentLogLevel)) {
        std::cout << message << std::endl;
    }
}

// Parse log level from an integer
LogLevel parseLogLevel(int level) {
    if (level < 0 || level > 3) {
        throw std::invalid_argument("Log level must be between 0 (Silent) and 3 (Verbose).");
    }
    return static_cast<LogLevel>(level);
}

// Initialize logger with a given numeric level
void initializeLogger(int logLevelValue) {
    try {
        LogLevel level = parseLogLevel(logLevelValue);
        setLogLevel(logLevelValue);
    } catch (const std::exception& e) {
        throw std::runtime_error("Error initializing logger: " + std::string(e.what()));
    }
}


    // Get current timestamp
    std::string getLocalTimestamp(){
    auto now = std::chrono::system_clock::now();
    std::time_t currentTime = std::chrono::system_clock::to_time_t(now);
    std::tm* localTime = std::localtime(&currentTime);
    std::ostringstream oss;
    oss <<std::put_time(localTime, "%Y-%m-%d %H:%M:%S");
    std::string datetime = oss.str();
    return datetime;
}

// Convert hex-String zu uint32_t
uint32_t hexStringToUint32(const std::string& hexStr) {
    uint32_t result;
    std::stringstream ss;
    ss << std::hex << hexStr;
    ss >> result;
    return result;
}

// Convert string to integer
int stringToInt(const std::string& str) {
    try {
        return std::stoi(str);
    } catch (const std::invalid_argument& e) {
        std::cerr << "Ungültiger String: Kann nicht in einen Integer konvertiert werden. Rückgabe: 0" << std::endl;
        return 0;
    } catch (const std::out_of_range& e) {
        std::cerr << "Wert außerhalb des gültigen Bereichs für einen Integer. Rückgabe: 0" << std::endl;
        return 0;
    }
}

// Convert string to float
float stringToFloat(const std::string& str) {
    try {
        return std::stof(str);
    } catch (const std::invalid_argument& e) {
        std::cerr << str << "Ungültiger String: Kann nicht in einen Float konvertiert werden. Rückgabe: 0.0" << std::endl;
        return 0.0f;
    } catch (const std::out_of_range& e) {
        std::cerr << "Wert außerhalb des gültigen Bereichs für einen Float. Rückgabe: 0.0" << std::endl;
        return 0.0f;
    }
}


int base64CharToValue(char c) {
    if (c >= 'A' && c <= 'Z') return c - 'A';
    if (c >= 'a' && c <= 'z') return c - 'a' + 26;
    if (c >= '0' && c <= '9') return c - '0' + 52;
    if (c == '+') return 62;
    if (c == '/') return 63;
    throw std::invalid_argument("Ungültiges Base64-Zeichen");
}

// Convert Base64 to Hex-Array
std::vector<uint8_t> base64ToHexArray(const std::string& base64String) {
    std::vector<uint8_t> hexArray;
    int val = 0;
    int bits = -8;
    
    for (const char& c : base64String) {
        if (c == '=') break;

        val = (val << 6) + base64CharToValue(c);
        bits += 6;

        if (bits >= 0) {
            uint8_t byte = (val >> bits) & 0xFF;
            hexArray.push_back(byte);
            bits -= 8;
        }
    }

    return hexArray;
}

// Function to retrieve the value of a specific key from JSON data
std::string getJsonValue(const std::string& jsonData, const std::string& keyName) {
    // Construct the search key
    std::string searchKey = "\"" + keyName + "\"";
    size_t keyPos = jsonData.find(searchKey);

    // Return empty string if key is not found
    if (keyPos == std::string::npos) {
        return "";
    }

    // Find the start of the value after the key
    size_t valueStart = jsonData.find(":", keyPos);
    if (valueStart == std::string::npos) {
        return "";
    }
    valueStart++;

    // Determine if the value is a string (enclosed in quotes), a number, or an array/object
    size_t valueEnd;
    std::string value;
    if (jsonData[valueStart] == '"') {
        // String value
        valueStart = jsonData.find('"', valueStart) + 1;
        valueEnd = jsonData.find('"', valueStart);
        if (valueEnd == std::string::npos) {
            return "";
        }
        value = jsonData.substr(valueStart, valueEnd - valueStart);
    } else if (jsonData[valueStart] == '[' || jsonData[valueStart] == '{') {
        // Array or object value
        char openBrace = jsonData[valueStart];
        char closeBrace = (openBrace == '[') ? ']' : '}';
        int braceCount = 1;
        valueEnd = valueStart + 1;

        // Find the matching closing brace
        while (valueEnd < jsonData.size() && braceCount > 0) {
            if (jsonData[valueEnd] == openBrace) braceCount++;
            else if (jsonData[valueEnd] == closeBrace) braceCount--;
            valueEnd++;
        }

        // Extract the entire array or object as a string
        value = jsonData.substr(valueStart, valueEnd - valueStart);
    } else {
        // Numeric or boolean value
        valueEnd = jsonData.find_first_of(",}", valueStart);
        if (valueEnd == std::string::npos) {
            return "";
        }
        value = jsonData.substr(valueStart, valueEnd - valueStart);
    }

    // Trim whitespace from value
    value.erase(0, value.find_first_not_of(" \n\r\t"));
    value.erase(value.find_last_not_of(" \n\r\t") + 1);

    return value;  // Return the extracted value
}


// List of lorawan network operator and NetIDs
std::string getNetId(const std::string& devaddr) {
    static const std::map<std::pair<uint32_t, uint32_t>, std::string> netIdMap = {
    {{0x000000, 0x01FFFFFF}, "Private/experimental nodes"},
    {{0x02000000, 0x03FFFFFF}, "Private/experimental nodes"},
    {{0x04000000, 0x05FFFFFF}, "Actility"},
    {{0x06000000, 0x07FFFFFF}, "Proximus"},
    {{0x08000000, 0x09FFFFFF}, "Swisscom"},
    {{0x0A000000, 0x0BFFFFFF}, "unassigned"},
    {{0x0C000000, 0x0DFFFFFF}, "unassigned"},
    {{0x0E000000, 0x0FFFFFFF}, "Bouygues Telecom"},
    {{0x10000000, 0x11FFFFFF}, "Orbiwise"},
    {{0x12000000, 0x13FFFFFF}, "SENET"},
    {{0x14000000, 0x15FFFFFF}, "KPN"},
    {{0x16000000, 0x17FFFFFF}, "EveryNet"},
    {{0x18000000, 0x19FFFFFF}, "unassigned"},
    {{0x1A000000, 0x1BFFFFFF}, "SK Telecom"},
    {{0x1C000000, 0x1DFFFFFF}, "SagemCom"},
    {{0x1E000000, 0x1FFFFFFF}, "Orange"},
    {{0x20000000, 0x21FFFFFF}, "A2A Smart City"},
    {{0x22000000, 0x23FFFFFF}, "unassigned"},
    {{0x24000000, 0x25FFFFFF}, "Kerlink"},
    {{0x26000000, 0x27FFFFFF}, "The Things Network"},
    {{0x28000000, 0x29FFFFFF}, "unassigned"},
    {{0x2A000000, 0x2BFFFFFF}, "Cisco Systems"},
    {{0x2C000000, 0x2DFFFFFF}, "unassigned"},
    {{0x2E000000, 0x2FFFFFFF}, "MultiTech Systems"},
    {{0x30000000, 0x31FFFFFF}, "Loriot"},
    {{0x32000000, 0x33FFFFFF}, "unassigned"},
    {{0x34000000, 0x35FFFFFF}, "unassigned"},
    {{0x36000000, 0x37FFFFFF}, "unassigned"},
    {{0x38000000, 0x39FFFFFF}, "unassigned"},
    {{0x3A000000, 0x3BFFFFFF}, "unassigned"},
    {{0x3C000000, 0x3DFFFFFF}, "unassigned"},
    {{0x3E000000, 0x3FFFFFFF}, "unassigned"},
    {{0x40000000, 0x41FFFFFF}, "unassigned"},
    {{0x42000000, 0x43FFFFFF}, "unassigned"},
    {{0x44000000, 0x45FFFFFF}, "Comcast"},
    {{0x46000000, 0x47FFFFFF}, "Ventia"},
    {{0x48000000, 0x49FFFFFF}, "unassigned"},
    {{0x4A000000, 0x4BFFFFFF}, "unassigned"},
    {{0x4C000000, 0x4DFFFFFF}, "unassigned"},
    {{0x4E000000, 0x4FFFFFFF}, "unassigned"},
    {{0x50000000, 0x51FFFFFF}, "unassigned"},
    {{0x52000000, 0x53FFFFFF}, "unassigned"},
    {{0x54000000, 0x55FFFFFF}, "unassigned"},
    {{0x56000000, 0x57FFFFFF}, "unassigned"},
    {{0x58000000, 0x59FFFFFF}, "unassigned"},
    {{0x5A000000, 0x5BFFFFFF}, "unassigned"},
    {{0x5C000000, 0x5DFFFFFF}, "unassigned"},
    {{0x5E000000, 0x5FFFFFFF}, "unassigned"},
    {{0x60000000, 0x61FFFFFF}, "SoftBank"},
    {{0x62000000, 0x63FFFFFF}, "unassigned"},
    {{0x64000000, 0x65FFFFFF}, "unassigned"},
    {{0x66000000, 0x67FFFFFF}, "unassigned"},
    {{0x68000000, 0x69FFFFFF}, "unassigned"},
    {{0x6A000000, 0x6BFFFFFF}, "Shenzhen Tencent Computer Systems Company Limited"},
    {{0x6C000000, 0x6DFFFFFF}, "Netze BW GmbH"},
    {{0x6E000000, 0x6FFFFFFF}, "Tektelic"},
    {{0x70000000, 0x71FFFFFF}, "Charter Communicaton"},
    {{0x72000000, 0x73FFFFFF}, "Amazon"},
    {{0x74000000, 0x75FFFFFF}, "Minol ZENNER Connect GmbH"},
    {{0x76000000, 0x77FFFFFF}, "Semtech"},
    {{0x78000000, 0x79FFFFFF}, "The Helium Foundation"},
    {{0x7A000000, 0x7BFFFFFF}, "Vorboss Limited"},
    {{0x7C000000, 0x7DFFFFFF}, "Unidata S.P.A."},
    {{0x7E000000, 0x7FFFFFFF}, "Birdz"},
    {{0xE0020000, 0xE003FFFF}, "Digita"},
    {{0xE0040000, 0xE005FFFF}, "Netmore"},
    {{0xE0060000, 0xE007FFFF}, "QuaeNet Inc."},
    {{0xE0080000, 0xE009FFFF}, "unassigned"},
    {{0xE00A0000, 0xE00BFFFF}, "IoT Network AS"},
    {{0xE00C0000, 0xE00DFFFF}, "unassigned"},
    {{0xE00E0000, 0xE00FFFFF}, "EDF"},
    {{0xE0100000, 0xE011FFFF}, "Unidata"},
    {{0xE0120000, 0xE013FFFF}, "unassigned"},
    {{0xE0140000, 0xE015FFFF}, "Oeresundskraft"},
    {{0xE0160000, 0xE017FFFF}, "unassigned"},
    {{0xE0180000, 0xE019FFFF}, "unassigned"},
    {{0xE01A0000, 0xE01BFFFF}, "unassigned"},
    {{0xE01C0000, 0xE01DFFFF}, "Spark"},
    {{0xE01E0000, 0xE01FFFFF}, "unassigned"},
    {{0xE0200000, 0xE021FFFF}, "Senet"},
    {{0xE0220000, 0xE023FFFF}, "unassigned"},
    {{0xE0240000, 0xE025FFFF}, "unassigned"},
    {{0xE0260000, 0xE027FFFF}, "Actility"},
    {{0xE0280000, 0xE029FFFF}, "Kerlink"},
    {{0xE02A0000, 0xE02BFFFF}, "unassigned"},
    {{0xE02C0000, 0xE02DFFFF}, "Cisco"},
    {{0xE02E0000, 0xE02FFFFF}, "Schneider Electric"},
    {{0xE0300000, 0xE031FFFF}, "Minol ZENNER Connect GmbH"},
    {{0xE0320000, 0xE033FFFF}, "unassigned"},
    {{0xE0340000, 0xE035FFFF}, "unassigned"},
    {{0xE0360000, 0xE037FFFF}, "Shenzhen Tencent Computer Systems Company Limited"},
    {{0xE0380000, 0xE039FFFF}, "MachineQ/Comcast"},
    {{0xE03A0000, 0xE03BFFFF}, "NTT (Nippon Telephone and Telegraph)"},
    {{0xE03C0000, 0xE03DFFFF}, "unassigned"},
    {{0xE03E0000, 0xE03FFFFF}, "KPN"},
    {{0xE0400000, 0xE041FFFF}, "Spectrum"},
    {{0xE0420000, 0xE043FFFF}, "Microshare Inc."},
    {{0xE0440000, 0xE045FFFF}, "unassigned"},
    {{0xE0460000, 0xE047FFFF}, "unassigned"},
    {{0xE0480000, 0xE049FFFF}, "Netze BW GmbH"},
    {{0xE04A0000, 0xE04BFFFF}, "Tektelic"},
    {{0xE04C0000, 0xE04DFFFF}, "unassigned"},
    {{0xE04E0000, 0xE04FFFFF}, "Birdz"},
    {{0xE0500000, 0xE051FFFF}, "Charter Communication"},
    {{0xE0520000, 0xE053FFFF}, "Machines Talk"},
    {{0xE0540000, 0xE055FFFF}, "Neptune Technology Group"},
    {{0xE0560000, 0xE057FFFF}, "Amazon"},
    {{0xE0580000, 0xE059FFFF}, "unassigned"},
    {{0xE05A0000, 0xE05BFFFF}, "Decentralized Wireless Foundation Inc"},
    {{0xE05C0000, 0xE05DFFFF}, "Eutelsat S.A."},
    {{0xE05E0000, 0xE05FFFFF}, "Meshify, Inc."},
    {{0xE0600000, 0xE061FFFF}, "EchoStar Mobile Ltd"},
    {{0xE0620000, 0xE063FFFF}, "unassigned"},
    {{0xE0640000, 0xE065FFFF}, "Orange"},
    {{0xE0660000, 0xE067FFFF}, "unassigned"},
    {{0xE0680000, 0xE069FFFF}, "Semtech"},
    {{0xE06A0000, 0xE06BFFFF}, "Wyld Networks"},
    {{0xE06C0000, 0xE06DFFFF}, "Unabiz"},
    {{0xE06E0000, 0xE06FFFFF}, "Vorboss Limited"},
    {{0xFC000000, 0xFC0003FF}, "RESERVED"},
    {{0xFC000400, 0xFC0007FF}, "unassigned"},
    {{0xFC000800, 0xFC000BFF}, "ResIOT"},
    {{0xFC000C00, 0xFC000FFF}, "SYSDEV"},
    {{0xFC001000, 0xFC0013FF}, "unassigned"},
    {{0xFC001400, 0xFC0017FF}, "unassigned"},
    {{0xFC001800, 0xFC001BFF}, "unassigned"},
    {{0xFC001C00, 0xFC001FFF}, "unassigned"},
    {{0xFC002000, 0xFC0023FF}, "Definium Technologies"},
    {{0xFC002400, 0xFC0027FF}, "unassigned"},
    {{0xFC002800, 0xFC002BFF}, "unassigned"},
    {{0xFC002C00, 0xFC002FFF}, "3S"},
    {{0xFC003000, 0xFC0033FF}, "unassigned"},
    {{0xFC003400, 0xFC0037FF}, "Packetworx"},
    {{0xFC003800, 0xFC003BFF}, "unassigned"},
    {{0xFC003C00, 0xFC003FFF}, "unassigned"},
    {{0xFC004000, 0xFC0043FF}, "unassigned"},
    {{0xFC004400, 0xFC0047FF}, "unassigned"},
    {{0xFC004800, 0xFC004BFF}, "Netmore"},
    {{0xFC004C00, 0xFC004FFF}, "Lyse AS"},
    {{0xFC005000, 0xFC0053FF}, "VTC Digicom"},
    {{0xFC005400, 0xFC0057FF}, "Machines Talk"},
    {{0xFC005800, 0xFC005BFF}, "Schneider Electric"},
    {{0xFC005C00, 0xFC005FFF}, "Connexin"},
    {{0xFC006000, 0xFC0063FF}, "Minol ZENNER Connect GmbH"},
    {{0xFC006400, 0xFC0067FF}, "Telekom Srbija"},
    {{0xFC006800, 0xFC006BFF}, "REQUEA"},
    {{0xFC006C00, 0xFC006FFF}, "Sensor Network Services"},
    {{0xFC007000, 0xFC0073FF}, "unassigned"},
    {{0xFC007400, 0xFC0077FF}, "Boston Networks Limited"},
    {{0xFC007800, 0xFC007BFF}, "unassigned"},
    {{0xFC007C00, 0xFC007FFF}, "mcf88 SRL"},
    {{0xFC008000, 0xFC0083FF}, "unassigned"},
    {{0xFC008400, 0xFC0087FF}, "Hiber"},
    {{0xFC008800, 0xFC008BFF}, "unassigned"},
    {{0xFC008C00, 0xFC008FFF}, "unassigned"},
    {{0xFC009000, 0xFC0093FF}, "NTT (Nippon Telephone and Telegraph)"},
    {{0xFC009400, 0xFC0097FF}, "unassigned"},
    {{0xFC009800, 0xFC009BFF}, "Mirakonta"},
    {{0xFC009C00, 0xFC009FFF}, "unassigned"},
    {{0xFC00A000, 0xFC00A3FF}, "Lacuna Space"},
    {{0xFC00A400, 0xFC00A7FF}, "Andorra Telecom"},
    {{0xFC00A800, 0xFC00ABFF}, "Xiamen Milesight IoT Co., Ltd."},
    {{0xFC00AC00, 0xFC00AFFF}, "Grenoble Alps University"},
    {{0xFC00B000, 0xFC00B3FF}, "unassigned"},
    {{0xFC00B400, 0xFC00B7FF}, "unassigned"},
    {{0xFC00B800, 0xFC00BBFF}, "Spectrum"},
    {{0xFC00BC00, 0xFC00BFFF}, "Afnic"},
    {{0xFC00C000, 0xFC00C3FF}, "unassigned"},
    {{0xFC00C400, 0xFC00C7FF}, "unassigned"},
    {{0xFC00C800, 0xFC00CBFF}, "Microshare Inc."},
    {{0xFC00CC00, 0xFC00CFFF}, "HEIG-VD"},
    {{0xFC00D000, 0xFC00D3FF}, "unassigned"},
    {{0xFC00D400, 0xFC00D7FF}, "unassigned"},
    {{0xFC00D800, 0xFC00DBFF}, "unassigned"},
    {{0xFC00DC00, 0xFC00DFFF}, "Alperia Fiber"},
    {{0xFC00E000, 0xFC00E3FF}, "First Snow Co., Ltd"},
    {{0xFC00E400, 0xFC00E7FF}, "unassigned"},
    {{0xFC00E800, 0xFC00EBFF}, "Vutility"},
    {{0xFC00EC00, 0xFC00EFFF}, "Meshed Pty Ltd"},
    {{0xFC00F000, 0xFC00F3FF}, "Birdz"},
    {{0xFC00F400, 0xFC00F7FF}, "Arthur D Riley & Co Ltd"},
    {{0xFC00F800, 0xFC00FBFF}, "Komro GmbH"},
    {{0xFC00FC00, 0xFC00FFFF}, "unassigned"},
    {{0xFC010000, 0xFC0103FF}, "Ceske Radiokomunikace"},
    {{0xFC010400, 0xFC0107FF}, "unassigned"},
    {{0xFC010800, 0xFC010BFF}, "MIOT Melita.io Technology GmbH"},
    {{0xFC010C00, 0xFC010FFF}, "PROESYS S.r.l."},
    {{0xFC011000, 0xFC0113FF}, "MeWe, Inc"},
    {{0xFC011400, 0xFC0117FF}, "Alpha-Omega Technology GmbH & Co. KG"},
    {{0xFC011800, 0xFC011BFF}, "SSE Contracting Limited t/a Mayflower Smart Control"},
    {{0xFC011C00, 0xFC011FFF}, "VEGA Grieshaber KG"},
    {{0xFC012000, 0xFC0123FF}, "unassigned"},
    {{0xFC012400, 0xFC0127FF}, "API-K"},
    {{0xFC012800, 0xFC012BFF}, "unassigned"},
    {{0xFC012C00, 0xFC012FFF}, "Nova Track Ltd"},
    {{0xFC013000, 0xFC0133FF}, "IMT Atlantique"},
    {{0xFC013400, 0xFC0137FF}, "Machines Talk"},
    {{0xFC013800, 0xFC013BFF}, "Yosensi Sp. z o.o."},
    {{0xFC013C00, 0xFC013FFF}, "The Saira 2.0, LLC dba The IoT Solutions"},
    {{0xFC014000, 0xFC0143FF}, "Neptune Technology Group"},
    {{0xFC014400, 0xFC0147FF}, "myDevices"},
    {{0xFC014800, 0xFC014BFF}, "Savoie Mont Blanc University"},
    {{0xFC014C00, 0xFC014FFF}, "Decentralized Wireless Foundation Inc"},
    {{0xFC015000, 0xFC0153FF}, "X-Telia Group Inc"},
    {{0xFC015400, 0xFC0157FF}, "unassigned"},
    {{0xFC015800, 0xFC015BFF}, "Eutelsat S.A."},
    {{0xFC015C00, 0xFC015FFF}, "Bejing Dingtek Technology Corp. Ltd"},
    {{0xFC016000, 0xFC0163FF}, "The Things Network Foundation"},
    {{0xFC016400, 0xFC0167FF}, "Quandify AB"},
    {{0xFC016800, 0xFC016BFF}, "Hutchison Drei Austria GmbH"},
    {{0xFC016C00, 0xFC016FFF}, "Agrology, PBC"},
    {{0xFC017000, 0xFC0173FF}, "mhascaro GmbH"},
    {{0xFC017400, 0xFC0177FF}, "Log5 Data"},
    {{0xFC017800, 0xFC017BFF}, "Citysens GmbH"},
    {{0xFC017C00, 0xFC017FFF}, "Wyld Networks"},
    {{0xFC018000, 0xFC0183FF}, "Meshify, Inc."},
    {{0xFC018400, 0xFC0187FF}, "EchoStar Mobile Ltd"},
    {{0xFC018800, 0xFC018BFF}, "unassigned"},
    {{0xFC018C00, 0xFC018FFF}, "Hello Space Systems Technology A.S."},
    {{0xFC019000, 0xFC0193FF}, "Wien Energie GmbH"},
    {{0xFC019400, 0xFC0197FF}, "unassigned"},
    {{0xFC019800, 0xFC019BFF}, "Stichting ThingsIX Foundation"},
    {{0xFC019C00, 0xFC019FFF}, "unassigned"},
    {{0xFC01A000, 0xFC01A3FF}, "ElektronikSystem i Umea AB (ELSYS)"},
    {{0xFC01A400, 0xFC01A7FF}, "UPLINK Network GmbH"},
    {{0xFC01A800, 0xFC01ABFF}, "Nynet Ltd"},
    {{0xFC01AC00, 0xFC01AFFF}, "ETA2U SRL"},
    {{0xFC01B000, 0xFC01B3FF}, "Institute Digital Engineering at Technical University Wurzburg-Schweinfurt"},
    {{0xFC01B400, 0xFC01B7FF}, "MClimate"},
    {{0xFC01B800, 0xFC01BBFF}, "Cibicom A/S"},
    {{0xFC01BC00, 0xFC01BFFF}, "Archipelagos Labs"},
    {{0xFC01C000, 0xFC01C3FF}, "Akenza AG"},
    {{0xFC01C400, 0xFC01C7FF}, "Ticae Telecom S.L."},
    {{0xFC01C800, 0xFC01CBFF}, "Unabiz"},
    {{0xFC01CC00, 0xFC01CFFF}, "SLY Inc."},
    {{0xFC01D000, 0xFC01D3FF}, "Swisscom"},
    {{0xFC01D400, 0xFC01D7FF}, "Nell Inc"},
    {{0xFC01D800, 0xFC01DBFF}, "Omantel"},
    {{0xFC01DC00, 0xFC01DFFF}, "Chirp Wireless"},
    {{0xFE000000, 0xFE0007FF}, "RESERVED"},
    {{0xFE000800, 0xFE000FFF}, "RESERVED"},
    {{0xFE001000, 0xFE0017FF}, "Techtenna b.v."},
    {{0xFE001800, 0xFE001FFF}, "LNX Solutions"},
    {{0xFE002000, 0xFE0027FF}, "Cometa Sas"},
    {{0xFE002800, 0xFE002FFF}, "Provincie Zeeland / Pollex"},
    {{0xFE003000, 0xFE0037FF}, "Intersaar GmbH"},
    {{0xFE003800, 0xFE003FFF}, "Wyld Networks"},
    {{0xFE004000, 0xFE0047FF}, "HeNet B.V."},
    {{0xFE004800, 0xFE004FFF}, "SkyNet IoT"},
    {{0xFE005000, 0xFE0057FF}, "Kanton Zug"},
    {{0xFE005800, 0xFE005FFF}, "Nebra Ltd"},
    {{0xFE006000, 0xFE0067FF}, "unassigned"},
    {{0xFE006800, 0xFE006FFF}, "Borough Smart"},
    {{0xFE007000, 0xFE0077FF}, "ISILINE s.r.l."},
    {{0xFE007800, 0xFE007FFF}, "Schippers Europe B.V."},
    {{0xFE008000, 0xFE0087FF}, "Meshed Pty Ltd"},
    {{0xFE008800, 0xFE008FFF}, "RemEX Technologies Limited"},
    {{0xFE009000, 0xFE0097FF}, "Actility"},
    {{0xFE009800, 0xFE009FFF}, "Provincie Zeeland / Pollex"},
    {{0xFE00A000, 0xFE00A7FF}, "Tracpac Ltd"},
    {{0xFE00A800, 0xFE00AFFF}, "Cometa Sas"},
    {{0xFE00B000, 0xFE00B7FF}, "VTM IoT"},
    {{0xFE00B800, 0xFE00BFFF}, "Autoskope GmbH"},
    {{0xFE00C000, 0xFE00C7FF}, "Clariane"},
    {{0xFE00C800, 0xFE00CFFF}, "Clariane"},
    {{0xFE00D000, 0xFE00D7FF}, "Sparknet Technologies"},
    {{0xFE00D800, 0xFE00DFFF}, "Actility"}
    };

    // Convert DevAddr from string to uint32_t
    uint32_t devaddrInt = hexStringToUint32(devaddr);

    // Search address range
    for (const auto& [range, operatorName] : netIdMap) {
        if (devaddrInt >= range.first && devaddrInt <= range.second) {
            return operatorName;
        }
    }

    return "unassigned";
}
