#include "network.h"

#include <iostream>
#include <string>
#include <cstring>
#include <iomanip>

#ifdef _WIN32
    #include <winsock2.h>
    #include <iphlpapi.h>
    #pragma comment(lib, "iphlpapi.lib")
    #pragma comment(lib, "Ws2_32.lib")
    #include <ws2tcpip.h>
#else
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <ifaddrs.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #include <unistd.h>
#endif

bool initializeNetwork() {
#ifdef _WIN32
    WSADATA wsaData;
    if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0) {
        std::cerr << "WinSock could not be initialized." << std::endl;
        return false;
    }
    std::cout << "WinSock successfully initialized." << std::endl;
#endif
    return true;
}

void printIPAddresses() {
#ifdef _WIN32
    ULONG bufferSize = 15000;
    PIP_ADAPTER_ADDRESSES adapterAddresses = (IP_ADAPTER_ADDRESSES*)malloc(bufferSize);
    if (GetAdaptersAddresses(AF_INET, GAA_FLAG_INCLUDE_PREFIX, nullptr, adapterAddresses, &bufferSize) == NO_ERROR) {
        for (PIP_ADAPTER_ADDRESSES adapter = adapterAddresses; adapter != nullptr; adapter = adapter->Next) {
            if (adapter->OperStatus == IfOperStatusUp) {
                for (PIP_ADAPTER_UNICAST_ADDRESS ua = adapter->FirstUnicastAddress; ua != nullptr; ua = ua->Next) {
                    char ipAddress[INET_ADDRSTRLEN];
                    sockaddr_in* sockaddr = (sockaddr_in*)ua->Address.lpSockaddr;
                    inet_ntop(AF_INET, &sockaddr->sin_addr, ipAddress, INET_ADDRSTRLEN);
                    if (std::string(ipAddress) != "127.0.0.1") {
                        std::cout << "IP Address: " << ipAddress << std::endl;
                    }
                }
            }
        }
    } else {
        std::cerr << "Could not retrieve adapter information." << std::endl;
    }
    free(adapterAddresses);
    WSACleanup();
#else
    struct ifaddrs *interfaces = nullptr;
    struct ifaddrs *ifa = nullptr;
    if (getifaddrs(&interfaces) == -1) {
        std::cerr << "Could not retrieve network adapters." << std::endl;
        return;
    }
    for (ifa = interfaces; ifa != nullptr; ifa = ifa->ifa_next) {
        if (ifa->ifa_addr && ifa->ifa_addr->sa_family == AF_INET) {
            char ipAddress[INET_ADDRSTRLEN];
            void* addr = &((struct sockaddr_in*)ifa->ifa_addr)->sin_addr;
            inet_ntop(AF_INET, addr, ipAddress, INET_ADDRSTRLEN);
            if (std::string(ipAddress) != "127.0.0.1") {
                std::cout << "IP Address: " << ipAddress << std::endl;
            }
        }
    }
    freeifaddrs(interfaces);
#endif
}

int initializeUDPListener(int& udpSocket, sockaddr_in& serverAddr) {
#ifdef _WIN32
    udpSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (udpSocket == INVALID_SOCKET) {
        std::cerr << "Error creating socket. Code: " << WSAGetLastError() << std::endl;
        return -1;
    }
#else
    udpSocket = socket(AF_INET, SOCK_DGRAM, 0);
    if (udpSocket < 0) {
        std::cerr << "Error creating socket." << std::endl;
        return -1;
    }
#endif

    std::memset(&serverAddr, 0, sizeof(serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_addr.s_addr = INADDR_ANY;
    serverAddr.sin_port = htons(1700);

#ifdef _WIN32
    if (bind(udpSocket, (sockaddr*)&serverAddr, sizeof(serverAddr)) == SOCKET_ERROR) {
        std::cerr << "Error binding socket to port 1700. Code: " << WSAGetLastError() << std::endl;
        closesocket(udpSocket);
        WSACleanup();
        return -1;
    }
#else
    if (bind(udpSocket, (sockaddr*)&serverAddr, sizeof(serverAddr)) < 0) {
        std::cerr << "Error binding socket to port 1700." << std::endl;
        close(udpSocket);
        return -1;
    }
#endif

    return 0;
}

void closeUDPListener(int udpSocket) {
#ifdef _WIN32
    closesocket(udpSocket);
    WSACleanup();
#else
    close(udpSocket);
#endif
}